-- iterator code by AlexanderMarinov
-- Compatible with Lua 5.1 (not 5.0).

-- NEEDS lfs LuaFileSystem (http://keplerproject.github.com/luafilesystem/)
-- NEEDS LPeg (http://www.inf.puc-rio.br/~roberto/lpeg.html)
require "lfs"
require "lexers"

  --[[prova1
       __va2
         __3]]
local a=1
if a==a then
 --[[prova1
      __va2
        __3]]
end

--- iterator on dir & files 
-- iterator on dir & files recursively
--@param dir path
--@param ext filter files extracted by extension (ex. .lua)
--@return iterator on files
-----------------------------------
function dirtree(dir, ext)
  assert(dir and dir ~= "", "directory parameter is missing or empty")
  if string.sub(dir, -1) == "/" then
    dir = string.sub(dir, 1, -2)
  end

  local diriters = {lfs.dir(dir)}
  local dirs = {dir}

  return function()
    repeat
    local entry = diriters[#diriters]()
    if entry then
      if entry ~= "." and entry ~= ".." then
        local filename = table.concat(dirs, "/").."/"..entry
        local attr = lfs.attributes(filename)
        if attr.mode == "directory" then
          table.insert(dirs, entry)
          table.insert(diriters, lfs.dir(filename))
        end
        if string.lower(string.sub(filename, -4)) == string.lower(ext) then
          return filename, attr
        end
      end
    else
      table.remove(dirs)
      table.remove(diriters)
    end
    until #diriters == 0
  end
end

--- string trim
--- trims a string
-- @param s string
-- @return trimmed string
function trim (s)
  return (string.gsub(s, "^%s*(.-)%s*$", "%1"))
end

--- string tab-trim
--- tab-trims a string
-- @param s string
-- @return trimmed string
function tabtrim (s)
  return (string.gsub(s, "^\t*(.-)\t*$", "%1"))
end

--- string rtrim
--- right-trims a string
-- @param s string
-- @return right-trimmed string
function rtrim (s)
  local n = #s
  while n > 0 and s:find("^%s", n) do
    n = n - 1
  end
  return s:sub(1, n)
end

--- string rtabtrim
--- right-tab-trims a string
-- @param s string
-- @return right-tab-trimmed string
function rtabtrim (s)
  local n = #s
  while n > 0 and s:find("^\t", n) do
    n = n - 1
  end
  return s:sub(1, n)
end

--- adds .lua extension to filename
-- adds .lua extension to filename
-- @param pfilename filename
-- @return filename with extension
function add_ext_lua(pfilename)
  if string.lower(string.sub(pfilename, -4)) ~= ".lua" then
    return pfilename..".lua"
  end
  return pfilename
end

--- cleans filename for dot charting
-- cleans filename for dot charting
-- @param pinitialdir initial directory
-- @param pfile filename
-- @return filename cleaned
function getfilenamefordot(pinitialdir, pfile)
  local name = pfile
  if name:find(pinitialdir) == 1 then
    name = string.gsub(name, pinitialdir, "", 1)
  end
  name = string.gsub(name, "\\", "/")
  if string.sub(name, 1, 1) == "/" then
    name = string.sub(name, 2)
  end
  name = string.gsub(name, ".LUA", ".lua")
  name = string.gsub(name, ".Lua", ".lua")
  return name
end

--- analyzes code, reformat code, writes .doc add LUADOC comment
-- analyzes code, reformat code, writes .doc add LUADOC comment
-- @param pinitialdir initial directory
-- @param pfile .lua filename
-- @param filegraph file for .dot chart
-- @param bTabForIndent enable tab for indenting
-- @param bWriteLex enable writing .lua.lex files
-- @param bNoReformat disable reformatting
-- @param bLuadoc enable writing LUADOC skeleton comment if missing
-- @param numspaces number of spaces used in indenting
function analyze_reformat(pinitialdir, pfile, filegraph, bTabForIndent, bWriteLex, bNoReformat, bLuadoc, numspaces, bIndentblockcomment)

  if filegraph then
    filegraph:write('"'..getfilenamefordot(pinitialdir, pfile)..'" [', "\n")
    filegraph:write('label = "{<f0> '..getfilenamefordot(pinitialdir, pfile).."||")
  end

  local fh = io.open(pfile, "r")

  linez = ""
  local line = ""
  while true do
    line = fh:read()
    if not line then
      break
    end
    --table.insert(linez,line)
    --print (line)
    linez = linez..line.."\n"
  end
  fh:close(fh)
  tokens = lexers(linez)

  local donotautoreformat = bNoReformat
  -- if exists <dir>/.do_ not_ autoreformat
  -- no auto reformat (default)
  filedonotautoreformat, msg = io.open(pinitialdir.."/.do_not_autoreformat", "r")
  if filedonotautoreformat ~= nil then
    donotautoreformat = true
  end

  local filelex
  if bWriteLex then
    filelex = assert(io.open(pfile..".lex", "w"))
  end

  local fileind = assert(io.open(pfile..".ind", "w"))

  local fileindstr = ""

  local oldtoken = nil
  local oldoldtoken = nil
  local print, format = _G.print, _G.string.format
  local oldline = 1
  local strlineold = ""
  local strline = ""
  local level = 0
  local levelindent = 0
  local tablevelindent = 0
  local levelspecial = {}
  local tablevelspecial = {}
  local leveltoken = {}
  levelspecial[0] = 0
  tablevelspecial[0] = 0
  leveltoken[0] = ""

  local links = ""
  local functions = 0
  local bfunction = false
  local functioname = ""
  local arguments = {}
  local requires = 0
  local brequire = false
  local requirename = ""
  local dofiles = 0
  local dofile = 0
  local bdofile = false
  local dofilename = ""
  local initialfunction = 0
  local initialcomment = 0
  local finalcomment = 0
  local bcomment = false
  local comment = ""
  local initialluadoccomment = 0
  local finalluadoccomment = 0
  local bluadoccomment = false
  local luadoccomment = ""

  for _, token in pairs(tokens) do
    if bWriteLex then
      filelex:write(format('line %i, %s: `%s`', token[3], token[1], token[2]), "\n")
    end
    --if oldline~=token[3] then
    --	fileind:write("\n")
    --end
    if string.lower(token[2]) == "end" or string.lower(token[2]) == ")" or string.lower(token[2]) == "}" or string.lower(token[2]) == "then" or string.lower(token[2]) == "else" or string.lower(token[2]) == "elseif" then
      level = level - 1
      levelindent = levelspecial[level]
      tablevelindent = tablevelspecial[level]
    end

    if token[1] == "comment" and not bNoReformat then
      if string.lower(token[2]):find("do_autoreformat") ~= nil then
        donotautoreformat = false
      elseif string.lower(token[2]):find("do_not_autoreformat") ~= nil then
        donotautoreformat = true
      end
    end

    if token[1] == "comment" then
      if not bluadoccomment then
        luadoccomment = ""
        if string.sub(trim(token[2]), 1, 3) == "---" then
          luadoccomment = token[2]
          bluadoccomment = true
          initialluadoccomment = string.len(fileindstr)
        end
      else
        luadoccomment = luadoccomment .. token[2]
      end
    elseif token[1] == "whitespace" then
      if bluadoccomment then
        luadoccomment = luadoccomment .. token[2]
      end
    elseif bluadoccomment then
      finalluadoccomment = string.len(fileindstr)
      bluadoccomment = false
    end

    if token[1] == "comment" then
      if not bcomment then
        comment = ""
        if string.sub(trim(token[2]), 1, 2) == "--" and string.sub(trim(token[2]), 1, 4) ~= "--[[" then
          comment = token[2]
          bcomment = true
          initialcomment = string.len(fileindstr)
        end
      else
        comment = comment .. token[2]
      end
    elseif token[1] == "whitespace" then
      if bcomment then
        comment = comment .. token[2]
      end
    elseif bcomment then
      finalcomment = string.len(fileindstr)
      bcomment = false
    end

    -- check and reformats indenting
    if donotautoreformat == false then
      if token[1] == "whitespace" then
        if token[2]:find("\n") ~= nil then
          fileindstr = fileindstr..strline.."\n"
          if token[2]:find("\n", token[2]:find("\n")+1) ~= nil then
            fileindstr = fileindstr.."\n"
          end
          strlineold = strline
          strline = ""
        else
          if string.sub(strline, -1) ~= " " then
            strline = strline.." "
          end
        end
      else
        if strline == "" then
        	local sindentstring = ""
          if tablevelindent > 0 then
            sindentstring = sindentstring..string.rep("\t", tablevelindent)
          end
          if levelindent > 0 then
            sindentstring = sindentstring..string.rep(" ", levelindent)
          end
          if token[1] == "comment" and trim(string.sub(token[2],1,4))=="--[[" then
            if bIndentblockcomment then
           	  strline = strline..sindentstring
            	if oldtoken ~= nil and oldtoken[1]=="whitespace" then
            		local tmpoldtoken = oldtoken[2]
            		while string.find(tmpoldtoken,"\n") do
            			tmpoldtoken = string.sub(tmpoldtoken,string.find(tmpoldtoken,"\n")+string.len("\n"))
            		end
            		if tmpoldtoken==nil then
                  tmpoldtoken=""
                end
            		token[2] = string.gsub(token[2], "\n"..tmpoldtoken, "\n"..sindentstring)
            	else
            	  token[2] = string.gsub(token[2], "\n", "\n"..sindentstring)
            	end
            else
            	if oldtoken ~= nil and oldtoken[1]=="whitespace" then
            		local tmpoldtoken = oldtoken[2]
            		while string.find(tmpoldtoken,"\n") do
            			tmpoldtoken = string.sub(tmpoldtoken,string.find(tmpoldtoken,"\n")+string.len("\n"))
            		end
            		if tmpoldtoken==nil then
                  tmpoldtoken=""
                end
                strline = string.sub(strline,1,string.len(strline)-string.len(tmpoldtoken))..tmpoldtoken
            	end
            end
          else
          	strline = strline..sindentstring
          end
        end
        --if string.len(strline)>1 then
        --	if string.sub(strline, -1) ~=" " then
        --	  if token[1]=="operator" then
        --		  strline=strline.." "
        --		end
        --  end
        --end
        if token[2] == "," or token[2] == "==" or token[2] == "~=" or token[2] == "<" or token[2] == "<=" or token[2] == ">" or token[2] == ">=" or token[2] == "=" then
          if string.sub(strline, -1) == " " then
            strline = string.sub(strline, 1, -2)
          end
          if token[2] == "," then
            strline = strline..token[2].." "
          else
            strline = strline.." "..token[2].." "
          end
        else
          local tmptoken = token[2]
          tmptoken = rtrim(tmptoken)
          tmptoken = rtabtrim(tmptoken)
          strline = strline..tmptoken
        end
      end
    else
      fileindstr = fileindstr..token[2]
    end

    --- getting requires
    if string.lower(token[2]) == "require" then
      brequire = true
      requirename = ""
      requires = requires+1
    else
      if brequire then
        if token[1] == "identifier" or token[1] == "string" then
          requirename = token[2]
          if (string.sub(requirename, 1, 1) == '"' and string.sub(requirename, -1, -1) == '"') or (string.sub(requirename, 1, 1) == "'" and string.sub(requirename, -1, -1) == "'") then
            requirename = string.sub(requirename, 2, -2)
          end
          requirename = add_ext_lua(trim(requirename))
          links = links..'"'..getfilenamefordot(pinitialdir, pfile)..'"->"'..getfilenamefordot(pinitialdir, requirename)..'" [label="require"]'.."\n"
          brequire = false
        end
      end
    end

    --- getting dofile
    if string.lower(token[2]) == "dofile" then
      bdofile = true
      dofilename = ""
      dofiles = dofiles+1
    else
      if bdofile then
        if token[1] == "identifier" or token[1] == "string" then
          dofilename = token[2]
          if (string.sub(dofilename, 1, 1) == '"' and string.sub(dofilename, -1, -1) == '"') or (string.sub(dofilename, 1, 1) == "'" and string.sub(dofilename, -1, -1) == "'") then
            dofilename = string.sub(dofilename, 2, -2)
          end
          dofilename = add_ext_lua(trim(dofilename))
          links = links..'"'..getfilenamefordot(pinitialdir, pfile)..'"->"'..getfilenamefordot(pinitialdir, dofilename)..'" [label="dofile"]'.."\n"
          bdofile = false
        end
      end
    end

    --- getting function and arguments (graphing & luadoc)
    if string.lower(token[2]) == "function" then
      bfunction = true
      if oldtoken ~= nil then
        if oldtoken[2] == "," or oldtoken[2] == "return" or oldtoken[2] == "(" then
          bfunction = false
        end
      end
      if oldoldtoken ~= nil then
        if oldoldtoken[2] == "," or oldoldtoken[2] == "return" or oldoldtoken[2] == "(" then
          bfunction = false
        end
      end
      initialfunction = string.len(fileindstr)
      if bfunction then
        functionname = ""
        arguments = {}
        functions = functions+1
        if luadoccomment ~= "" then
          if finalluadoccomment < string.len(fileindstr)-10-string.len(strline) then
            luadoccomment = ""
            initialluadoccomment = 0
            finalluadoccomment = 0
            bluadoccomment = false
          end
        end
        if comment ~= "" then
          if finalcomment < string.len(fileindstr)-10-string.len(strline) then
            comment = ""
            initialcomment = 0
            finalcomment = 0
            bcomment = false
          end
        end
      end
    else
      if bfunction then
        if token[1] == "identifier" then
          if functionname == "" then
            functionname = token[2]
          else
            table.insert(arguments, token[2])
          end
        end
        if token[2] == ")" and string.lower(leveltoken[level]) == "function" then
          if bLuadoc and functionname ~= "" and functionname ~= "love.conf" then
            if luadoccomment ~= "" then
              commentluadoc = luadoccomment
              for karg, varg in pairs(arguments) do
                if string.lower(commentluadoc):find(string.lower("@param "..varg)) == nil then
                  if string.lower(commentluadoc):find(string.lower("\n-- @return")) == nil then
                    commentluadoc = commentluadoc.."-- @param "..varg.." (LUADOC TODO add "..varg.." description) \n"
                  else
                    local commentluadocini = string.sub(commentluadoc, 1, string.lower(commentluadoc):find(string.lower("-- @return"))-1)
                    local commentluadocfin = string.sub(commentluadoc, string.lower(commentluadoc):find(string.lower("-- @return")), string.len(commentluadoc))
                    commentluadoc = commentluadocini.."-- @param "..varg.." (LUADOC TODO add "..varg.." description) \n"..commentluadocfin
                  end
                end
              end
              --if string.lower(commentluadoc):find(string.lower("@return")) == nil then
              --  commentluadoc = commentluadoc.."-- @return".."\n"
              --end
              fileindstr = string.sub(fileindstr, 1, initialluadoccomment)..commentluadoc..string.sub(fileindstr, finalluadoccomment+1)
            elseif comment ~= "" then
              comment = trim(comment)
              if string.sub(comment, 1, 2) == "--" then
                comment = string.sub(comment, -string.len(comment)+2)
              end
              if string.sub(comment, -1, -1) == "\n" then
                comment = string.sub(comment, 1, string.len(comment)-1)
              end
              comment = trim(comment)
              local commentluadoc = "--- "..comment.." (LUADOC TODO check description)".."\n"
              for karg, varg in pairs(arguments) do
                commentluadoc = commentluadoc.."-- @param "..varg.."\n"
              end
              commentluadoc = commentluadoc.. "-- @return".."\n"
              fileindstr = string.sub(fileindstr, 1, initialcomment)..commentluadoc..string.sub(fileindstr, finalcomment+1)
            else
              local commentluadoc = "--- "..functionname.." (LUADOC TODO add resume)".."\n"
              commentluadoc = commentluadoc.."-- "..functionname.." (LUADOC TODO add description)".."\n"
              for karg, varg in pairs(arguments) do
                commentluadoc = commentluadoc.."-- @param "..varg.." (LUADOC TODO add "..varg.." description) \n"
              end
              commentluadoc = commentluadoc.. "-- @return".." (LUADOC TODO add return description) \n"
              fileindstr = string.sub(fileindstr, 1, initialfunction)..commentluadoc..string.sub(fileindstr, initialfunction+1)
            end
          end

          if filegraph and functionname ~= "" then
            filegraph:write(" +"..getfilenamefordot(pinitialdir, functionname)..'(')
            local num = 0
            for karg, varg in pairs(arguments) do
              if num > 0 then
                filegraph:write(",")
              end
              filegraph:write(varg)
              num = num + 1
            end
            filegraph:write(')\\l')
          end
          bfunction = false
        end
      end
    end

    if string.lower(token[2]) == "do" or string.lower(token[2]) == "(" or string.lower(token[2]) == "{" or string.lower(token[2]) == "then" or string.lower(token[2]) == "function" or string.lower(token[2]) == "if" or string.lower(token[2]) == "else" or string.lower(token[2]) == "elseif" then
      level = level + 1
      if string.lower(token[2]) == "(" or string.lower(token[2]) == "{" then
        if bTabForIndent then
          levelspecial[level] = string.len(strline)+string.len(token[2])-level+1
          levelindent = levelspecial[level]
          tablevelspecial[level] = level-1
          tablevelindent = tablevelspecial[level]
        else
          levelspecial[level] = string.len(strline)+string.len(token[2])
          levelindent = levelspecial[level]
          tablevelspecial[level] = 0
          tablevelindent = tablevelspecial[level]
        end
        leveltoken[level] = token[2]
      else
        if bTabForIndent then
          levelspecial[level] = 0
          levelindent = levelspecial[level]
          tablevelspecial[level] = level
          tablevelindent = tablevelspecial[level]
        else
          levelspecial[level] = level*numspaces
          levelindent = levelspecial[level]
          tablevelspecial[level] = 0
          tablevelindent = tablevelspecial[level]
        end
        leveltoken[level] = token[2]
      end
    end

    oldline = token[3]
    oldoldtoken = oldtoken
    oldtoken = token
  end

  fileindstr = fileindstr..strline

  fileind:write(fileindstr)

  if bWriteLex then
    filelex:write(format('total of %i tokens, %i lines', #tokens, tokens[#tokens][3]))
    filelex:close()
  end

  fileind:close()

  if filegraph then
    filegraph:write('}" shape="record"];', "\n")
    filegraph:write(links)
  end

end

--- print version
-- print_version 
local function print_version ()
  print ("Prettifier.lua".." Ver.".."0.0.2")
  print ("written by Bartoleo (http://bitbucket.org/bartoleo/prettydoclua/)")
  print ("for LOVE community (http://love2d.org/) and also LUA community")
  print ("  iterator code by AlexanderMarinov")
  print ("  lexer code by Peter Odding, 2007/04/04")
  print ("  using LPeg by Roberto Ierusalimschy")
end

--- print help/usage for command
-- print help/usage for command
-- @param error_argument argument not recognized
local function print_help (error_argument)
  print ("Usage: ".."prettifier.lua"..[[ root_dir [options]
Reindent/reformat .lua files.
Adding LUADOC skeleton comment if missing. (with --luadoc)
Creating .dot (uml charting) in root_dir/doc if present (with --chart)
For each .lua file will create a .lua.ind file containg reformat/LUADOC

 Available options are:
  --usetab               use tabs for indenting (default is spaces)
  --spaces:N             to specify the number of spaces indenting (default:2)
  --indentblockcomment   enable indenting of block comments (comments 
                         starting with ]].."--[["..[[ (experimental)
  --chart                create .dot file in root_dir/doc if present
  --noreformat           disable indenting/reformatting
  --luadoc               adds luadoc skeleton comment/params if missing
  --lex                  for each .lua file creates a .lua.lex containg
                         info on code tokenized
  --help                 print this help and exit
  --version              print version information]])
  print("")
  print("if you don't want to reformat any files in a dir you can put a file called")
  print("\".do_not_autoreformat\" in dir ")
  print("or in code if you want to disable reformat you can add this comment:")
  print("--do_not_autoreformat")
  print("and to reenable reformat")
  print("--do_autoreformat")
  print("")
  print_version()
  if error_argument ~= "" then
    print ("-----------------------------------------------")
    print ("Error : Argument " .. error_argument .. " not recognized, exiting...")
  end
end

local bTabForIndent = false
local bChart = false
local bWriteLex = false
local bNoReformat = false
local bLuadoc = false
local bIndentblockcomment = false
local numspaces = 2

-- command arguments
local initialdir = arg[1]
if initialdir == nil then
  initialdir = "."
end

local bHelp = false
local argc = #arg
if argc <= 0 then
  bHelp = true
end
local argn = 1
while argn <= #arg do
  local argi = arg[argn]
  if argi == "--usetab" then
    bTabForIndent = true
  elseif string.sub(argi,1,9) == "--spaces:" then
  	numspaces = string.sub(argi,10,12)
  	if numspaces ==""then
      print_help(argi)
      return false
    end
  elseif argi == "--indentblockcomment" then
    bIndentblockcomment = true
  elseif argi == "--chart" then
    bChart = true
  elseif argi == "--noreformat" then
    bNoReformat = true
  elseif argi == "--luadoc" then
    bLuadoc = true
  elseif argi == "--lex" then
    bWriteLex = true
  elseif argi == "--version" then
    print_version()
    return true
  elseif argi == "--help" or argi == "/?" or argi == "-help" or argi == "-?" then
    bHelp = true
  elseif argi == "/?" then
    bHelp = true
  else
    if argn > 1 or string.sub(argi, 1, 1) == "-" then
      print_help(argi)
      return false
    end
  end
  argn = argn + 1
end

if bHelp then
  print_help("")
  return true
end

local chartfilename
local filegraph
if bChart then
  if string.sub(initialdir, -1) ~= "\\" and string.sub(initialdir, -1) ~= "/" then
    chartfilename = initialdir.."/doc/chart.dot"
  else
    chartfilename = initialdir.."doc/chart.dot"
  end
  filegraph = io.open(chartfilename, "w")
end

if filegraph then
  filegraph:write('digraph relations {', "\n")
  filegraph:write('node [shape = rectangle];', "\n")
end

files = {}
for filename, attr in dirtree(initialdir, ".lua") do
  table.insert(files, filename)
end
--table.sort(files)
for kk, vv in pairs(files) do
  analyze_reformat(initialdir, vv, filegraph, bTabForIndent, bWriteLex, bNoReformat, bLuadoc,numspaces,bIndentblockcomment)
end
if filegraph then
  filegraph:write('}', "\n")
  filegraph:close()
end
